(*
 * Copyright (c) 2016 Jeremy Yallop.
 *
 * This file is distributed under the terms of the MIT License.
 * See the file LICENSE for details.
 *)

external setup_thread_registration : unit -> unit
  = "ctypes_setup_thread_registration"
